/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you add "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Smartphone garage door opener with web interface (aka iPhone garage door opener).
 * The switch which senses the state of the door is conneced
 * between GND and PC0..PC3. The door relays are controlled via PD7..PD4
 * This software supports 1 to 4 garage doors
 * 
 * See http://tuxgraphics.org/electronics/
 *
 * Chip type: Atmega328  or atmega644 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"
#include "websrv_help_functions.h"

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)
//
#define RELAYON0 PORTD|=(1<<PORTD7)
#define RELAYON1 PORTD|=(1<<PORTD6)
#define RELAYON2 PORTD|=(1<<PORTD5)
#define RELAYON3 PORTD|=(1<<PORTD4)
//#define ALLRELAYOFF PORTD&=~(1<<PORTD7)|(1<<PORTD6)|(1<<PORTD5)|(1<<PORTD4)
#define ALLRELAYOFF PORTD&=~0xf0

// please modify the following lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices. The IP address may be changed at run-time but the
// MAC address is hardcoded:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
//
// The IP of this device (can also be changed at run-time, see README file):
static uint8_t myip[4] = {10,0,0,29};
// listen port for tcp/www:
static uint16_t mywwwport=80;
static uint8_t maxdoornum=2; // The number of doors you have. Range: 1..4
//
// The following are initinal passwords. They can be changed during run-time
// and they are stored persistant in the eeprom. The master password (km) can
// be reset to compile time values by connecting pin PD2 to gnd at power on.
// A compile time password shorter than 2 characters is invalid. 
// In the page that allows you to change at run time the passwords an empty
// field does not change the password. Here in the source code the meaning is different.
// An empty string means passcode not pre-set.
static char c1[10]=""; // passcode 1, empty string, not set, max 9 characters long, can be any string
static char c2[10]=""; // passcode 2, max 9 characters long, not set, can be any string
static char km[10]="secret"; // master password, max 9 characters long, the master pw can be used to change the other passwords at run-time.
//
// -------------- do not change anything below this line ----------
//
// The buffer is the packet size we can handle and its upper limit is
// given by the amount of RAM that the atmega chip has. 
#define BUFFER_SIZE 950
static uint8_t buf[BUFFER_SIZE+1];
// global string buffer
#define STR_BUFFER_SIZE 40
static char gStrbuf[STR_BUFFER_SIZE+1];
//
static char browser_cookie[3][9];
static uint8_t browser_fingerprint[3]={0,0,0}; // zero means no entry
static uint8_t browser_cookie_attempt=1;
static uint16_t gPlen;
static uint8_t int_random=4; // small random number, changes all the time
static uint8_t random_arr[3]={7,7,7}; // here we store every minute a random number and step the index, we use this to encode key code urls
static uint8_t random_arr_index=0;
static uint8_t prev_doorstates=0; // the previous value of the lower 4 
static volatile uint8_t gsec=0; // seconds timer
static uint8_t min_high=0;
static uint16_t min_low=0;
static volatile uint8_t pushbuttontimer=0; // count down timer to switch relay off
static volatile uint8_t debounce=0; // debounce push buttons
// -------------- history storage
// those defines are more for documentation purposes, we use
// below some formulas to generate the same values.
#define DOOR0_OPEN 0
#define DOOR1_OPEN 0x10
#define DOOR2_OPEN 0x20
#define DOOR3_OPEN 0x30
#define DOOR0_CLOSE 1
#define DOOR1_CLOSE 0x11
#define DOOR2_CLOSE 0x21
#define DOOR3_CLOSE 0x31
#define DOOR0_C1 2
#define DOOR1_C1 0x12
#define DOOR2_C1 0x22
#define DOOR3_C1 0x32
#define DOOR0_C2 3
#define DOOR1_C2 0x13
#define DOOR2_C2 0x23
#define DOOR3_C2 0x33
#define DOOR0_Cfree0 4
#define DOOR1_Cfree0 0x14
#define DOOR2_Cfree0 0x24
#define DOOR3_Cfree0 0x34
#define KM 7
#define ENDMARKER 0xff
//
// We need need quite a bit of storage for the history. We use the eeprom
// area as this is slowly changing data and most chips have very little RAM
// How deep the history is (be careful a too big history will produce a too big web page for the network buffer to take):
#define HIST_BUFFER_SIZE 50
// position at which the EEPROM is used for temporary storage of data:
#define EEPROM_HIST_POS_OFFSET 150
// EEPROM storage is non volatile but we are storing time in a relative
// format. Time is relative to power on of the board. Therefore
// we can only use it as a temporary storage thing and have to invalidate
// the data after power on.
#define EEPROM_USERDAT_POS_OFFSET 50
static int8_t hist_pos=-1; // last position hist data was written, -1 no data written yet
// 
int8_t read_hist(int8_t pos,uint8_t *min_h,uint16_t *min_l,uint8_t *eventdata)
{
        if (pos<0){
                return(0); // no data
        }
        *min_h=(uint8_t)eeprom_read_byte((uint8_t *)(EEPROM_HIST_POS_OFFSET+(pos*4)));
        *min_l=(uint16_t)eeprom_read_word((uint16_t *)(EEPROM_HIST_POS_OFFSET+(pos*4)+1));
        *eventdata=(uint8_t)eeprom_read_byte((uint8_t *)(EEPROM_HIST_POS_OFFSET+(pos*4)+3));
        return(1); 
}

// no overflow protection here make sure you stay within the available eeprom size, we need (HIST_BUFFER_SIZE+1)*4 + EEPROM_HIST_POS_OFFSET bytes of eeprom:
void store_hist(uint8_t min_h,uint16_t min_l,uint8_t eventdata)
{
        hist_pos++;
        if (hist_pos>HIST_BUFFER_SIZE){
                hist_pos=0;
        }
        eeprom_write_byte((uint8_t *)(EEPROM_HIST_POS_OFFSET+(hist_pos*4)),min_h); 
        eeprom_write_word((uint16_t *)(EEPROM_HIST_POS_OFFSET+(hist_pos*4)+1),min_l); 
        eeprom_write_byte((uint8_t *)(EEPROM_HIST_POS_OFFSET+(hist_pos*4)+3),eventdata);
}

void set_hist_endmarker(void)
{
        int8_t pos=HIST_BUFFER_SIZE; 
        eeprom_write_byte((uint8_t *)(EEPROM_HIST_POS_OFFSET+(pos*4)+3),0xff);
}

// form EEPROM_USERDAT_POS_OFFSET onwards we store passwords and 
// browser cookie data. We put a magic number before each variable that we
// store this way we know if the variable was stored or not.
#define EE_MAGIC 43
void eeprom_store_browser_cookie(void)
{
        int8_t i=0;
        uint16_t pos;
        pos=EEPROM_USERDAT_POS_OFFSET+1;
        while(i<3){
                eeprom_write_block((void *)browser_cookie[i],(void *)pos,sizeof(browser_cookie[i]));
                pos+=sizeof(browser_cookie[i])+1;
                i++;
        }
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET,EE_MAGIC); // magic number
}

void eeprom_read_browser_cookie_conditional(void)
{
        int8_t i=0;
        uint16_t pos=EEPROM_USERDAT_POS_OFFSET+1;
        if (eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET)==EE_MAGIC){
                while(i<3){
                        eeprom_read_block((void *)browser_cookie[i],(void *)pos,sizeof(browser_cookie[i]));
                        pos+=sizeof(browser_cookie[i])+1;
                        browser_cookie[i][sizeof(browser_cookie[i])-1]='\0';
                        i++;
                }
        }
}

// start from EEPROM_USERDAT_POS_OFFSET + 40 to leave some extra space
void eeprom_store_browser_fingerprint(void)
{
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+41,browser_fingerprint[0]);
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+42,browser_fingerprint[1]);
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+43,browser_fingerprint[2]);
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+40,EE_MAGIC); // magic number
}

void eeprom_read_browser_fingerprint_conditional(void)
{
        if (eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+40)==EE_MAGIC){
                browser_fingerprint[0]=eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+41);
                browser_fingerprint[1]=eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+42);
                browser_fingerprint[2]=eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+43);
        }
}

// start from EEPROM_USERDAT_POS_OFFSET + 45 
void eeprom_store_passwords(void)
{
        uint16_t pos=EEPROM_USERDAT_POS_OFFSET+1+45;
        eeprom_write_block((void *)c1,(void *)pos,sizeof(c1));
        pos+=sizeof(c1)+1;
        eeprom_write_block((void *)c2,(void *)pos,sizeof(c2));
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+45,EE_MAGIC); // magic number
}

void eeprom_read_passwords_condtional(void)
{
        uint16_t pos=EEPROM_USERDAT_POS_OFFSET+1+45;
        if (eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+45)==EE_MAGIC){
                eeprom_read_block((void *)c1,(void *)pos,sizeof(c1));
                c1[sizeof(c1)-1]='\0';
                pos+=sizeof(c1)+1;
                eeprom_read_block((void *)c2,(void *)pos,sizeof(c2));
                c2[sizeof(c2)-1]='\0';
        }
}

// start from EEPROM_USERDAT_POS_OFFSET + 75
void eeprom_store_master(void)
{
        eeprom_write_block((void *)km,(void *)EEPROM_USERDAT_POS_OFFSET+75+1,sizeof(km));
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+75,EE_MAGIC); // magic number
}

void eeprom_invalidate_master(void)
{
        eeprom_write_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+75,1); // erase magic number
}

void eeprom_read_master_conditional(void)
{
        if (eeprom_read_byte((uint8_t *)EEPROM_USERDAT_POS_OFFSET+75)==EE_MAGIC){
                eeprom_read_block((void *)km,(void *)EEPROM_USERDAT_POS_OFFSET+75+1,sizeof(c1));
                km[sizeof(km)-1]='\0';
        }
}

uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}

uint16_t http200okjs(int8_t cache)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: application/x-javascript\r\n"));
        if (cache){
                return(fill_tcp_data_p(buf,plen,PSTR("\r\n")));
        }
        return(fill_tcp_data_p(buf,plen,PSTR("Pragma: no-cache\r\n\r\n")));
}

uint16_t fill_tcp_data_uint8_pad(uint8_t *buf,uint16_t plen,uint8_t i)
{
        itoa(i,gStrbuf,10); // convert integer to string
        if (i<10) plen=fill_tcp_data(buf,plen,"0");
        return(fill_tcp_data(buf,plen,gStrbuf));
}

uint16_t fill_tcp_data_int(uint8_t *buf,uint16_t plen,uint16_t ui)
{
        uint8_t i=6;
        if (ui==0){
                return(fill_tcp_data(buf,plen,"0"));
        }
        // convert a 16bit unsigned integer into a printable string, avr-lib has
        // not function for that therefore we do it step by step:
        gStrbuf[i]='\0';
        while(ui&&i){
                i--;
                gStrbuf[i]=(char)((ui%10 & 0xf))+0x30;
                ui=ui/10;
        }
        return(fill_tcp_data(buf,plen,&gStrbuf[i]));
}

// p1.js
uint16_t print_p1_js(void)
{
        // I am sorry that the javascript is a bit unreadable 
        // but we absolutely need to keep the amount of code
        // in bytes small.
        //
        uint16_t plen;
        if (random_arr[random_arr_index]==0){
                random_arr[random_arr_index]=43; // we can not use 0
        }
        plen=http200okjs(0);
        plen=fill_tcp_data_p(buf,plen,PSTR("function pad(s){if(s.length<2){return('0'+s)}return(s)}\n"));
        // the function cy makes it impossible to re-use the same URL-get more than 60sec later
        plen=fill_tcp_data_p(buf,plen,PSTR("function cr(f){ c="));
        itoa(random_arr[random_arr_index],gStrbuf,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(";r='")); // the first digit in the return value is the random_arr_index
        itoa(random_arr_index,gStrbuf,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("';s=f.value;i=s.length;j=0;while(j<i){r+=pad(((0xff & s.charCodeAt(j))^c).toString(16));j++} f.value='';return(r)}\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("function ni(d,n,v){var i=document.createElement('input');i.setAttribute('type','hidden'); i.setAttribute('name',n);i.setAttribute('value',v);d.insertBefore(i,d.firstChild)}\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("function cy(d){d.y.value=cr(d.c)}\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("function csm(d){ni(d,'y',cr(d.c));ni(d,'z',cr(d.n))}\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("function csp(d){ni(d,'y',cr(d.c));ni(d,'z1',cr(d.c1));ni(d,'z2',cr(d.c2))}\n"));
        return(plen);
}

// p2.js
uint16_t print_p2_js(void)
{
        // I am sorry that the javascript is a bit unreadable 
        // but we absolutely need to keep the amount of code
        // in bytes small.
        //
        // In this function we decode the history data.
        uint16_t plen;
        plen=http200okjs(1);
        plen=fill_tcp_data_p(buf,plen,PSTR("\
var t=new Date();\n\
function pad(n){var s=n.toString(10);if(s.length<2){return('0'+s)}return(s)}\n\
function dw(s){document.writeln(s)}\n\
function v2s(v){ var sv=Array('opened','closed','code 1','code 2','cfree0','cfree1','cfree2'); if (v==7) return('---- - master'); return('door '+(v>>4)+' '+sv[v&0xf]) }\n\
function pt(d){ return(d.getFullYear()+'-'+pad(1+d.getMonth())+'-'+pad(d.getDate())+' '+pad(d.getHours())+':'+pad(d.getMinutes()))}\n\
function dh(dt,dv){ var i,ms,l=new Date(); ms=t.getTime(); for(i=0;i<dv.length;i++){ l.setTime(ms-(dt[i]*60*1000)); dw(pt(l)+' = '+v2s(dv[i])); } dw('+++'); dw('current time: '+pt(t)) }\n\
"));
        return(plen);
}

uint16_t print_hist_array(uint16_t plen)
{
        // In this function we provide the history data as javascript arrays
        int8_t p=hist_pos;
        int8_t i=1;
        uint8_t min_h,eventdata;
        uint16_t min_l;
        if (p==-1){
                plen=fill_tcp_data_p(buf,plen,PSTR("gt=Array();gv=Array();"));
                return(plen);
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("gt=["));
        while(1){
                read_hist(p,&min_h,&min_l,&eventdata);
                if (eventdata==ENDMARKER){
                        break;
                }
                // Take the difference to current time, we can assume
                // that the timer never wraps as the wrap time is 31 years.
                // This circuit will not be powered on for straight 31 years.
                if (min_low<min_l){
                        // we must not overrun 16 bit
                        min_l=min_l-min_low;
                        min_l=(min_l ^ 0xffff); // subtract from 0xffff+1 is one's complement + 1
                        min_l+=1;
                        min_h=min_high-min_h-1;
                }else{
                        min_l=min_low-min_l;
                        min_h=min_high-min_h;
                }
                if (i>1){
                        plen=fill_tcp_data(buf,plen,","); // not the first time
                }
                i++;
                plen=fill_tcp_data_p(buf,plen,PSTR("0x"));
                itoa(min_h,gStrbuf,16); // convert integer to string
                plen=fill_tcp_data(buf,plen,gStrbuf);
                itoa(min_l,gStrbuf,16); // convert integer to string
                plen=fill_tcp_data(buf,plen,gStrbuf);
                p--;
                if (p<0){
                        p=HIST_BUFFER_SIZE;
                }
                if (i>HIST_BUFFER_SIZE){
                        // read all data once
                        break;
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("];\n"));
        i=1;
        p=hist_pos;
        plen=fill_tcp_data_p(buf,plen,PSTR("gv=["));
        while(1){
                read_hist(p,&min_h,&min_l,&eventdata);
                if (eventdata==ENDMARKER){
                        break;
                }
                if (i>1){
                        plen=fill_tcp_data(buf,plen,","); // not the first time
                }
                i++;
                plen=fill_tcp_data_p(buf,plen,PSTR("0x"));
                itoa(eventdata,gStrbuf,16); // convert integer to string
                plen=fill_tcp_data(buf,plen,gStrbuf);
                p--;
                if (p<0){
                        p=HIST_BUFFER_SIZE;
                }
                if (i>HIST_BUFFER_SIZE){
                        // read all data once
                        break;
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("];\n"));
        return(plen);
}


uint16_t print_line_mobile_webpage(uint16_t plen)
{
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        return(plen);
}

uint16_t print_webpage_ok_pw_update(char *str)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK, "));
        plen=fill_tcp_data(buf,plen,str);
        plen=fill_tcp_data_p(buf,plen,PSTR(" password(s) changed</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/ method=get><input type=submit value=\"continue\"></form>\n"));
        return(plen);
}

uint16_t print_webpage_ok(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/ method=get><input type=submit value=\"continue\"></form>\n"));
        return(plen);
}

uint16_t print_webpage_door_activatation(int8_t door)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK, activating door "));
        itoa(door,gStrbuf,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/ method=get><input type=submit value=\"continue\"></form>\n"));
        return(plen);
}

// key less access page, this will set in the next step a cookie. The
// expriry date of that cookie is given by "var x". x is in units of days.
uint16_t print_webpage_nokey(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a><br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=/p1.js></script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Enable cookie based access</h2>\n")); 
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>Passcode free access via your device.\n")); 
        plen=fill_tcp_data_p(buf,plen,PSTR("<form name=cf action=kk method=get onsubmit=cy(this)>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<input type=hidden name=y>\
<input type=hidden name=e>\
Store as device number: <select size=1 name=n><option>0<option>1<option>2</select>\n\n\
master pw: <input type=password name=c size=9> \
<input type=submit value=Enable></form>\n\
<script>\
var x=600;var d=new Date();\n\
d.setTime(d.getTime()+(x*24*3600*1000));\n\
document.cf.e.value=d.toGMTString();\n\
</script>\
<hr>\n\
"));
        return(plen);
}

uint16_t print_webpage_setcookie(char *expriry,char *browser_c)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("Set-Cookie: dk="));
        plen=fill_tcp_data(buf,plen,browser_c);
        plen=fill_tcp_data_p(buf,plen,PSTR("; expires="));
        plen=fill_tcp_data(buf,plen,expriry);
        plen=fill_tcp_data_p(buf,plen,PSTR("; path=/\r\n\r\n"));
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK, browser cookie set</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>continue...</a><br><hr>\n"));
        return(plen);
}

// this the history page
uint16_t print_webpage_hh(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=p2.js></script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>History info</h2>\n<pre>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script>\n"));
        plen=print_hist_array(plen); // for security reasons we must not allow access to this data from outside this page. It can not be an own .js file.
        plen=fill_tcp_data_p(buf,plen,PSTR("dh(gt,gv);\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("</script>\n<hr>\n"));
        return(plen);
}

// use the master key to change the passwords 1 or 2
uint16_t print_webpage_changekeys(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=/p1.js></script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<h2>Change passcodes</h2>\n\
<pre>\
<form action=sp method=get onsubmit=csp(this)>\n\
passcode1:<input type=text name=c1>\n\
passcode2:<input type=text name=c2>\n\
master pw:<input type=password name=c> \
<input type=submit value=change></form></pre>\n\
"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<h2>Clear all key free access</h2>\n\
<pre>\
<form action=dd method=get onsubmit=\"ni(this,'y',cr(this.c));\">\n\
master pw:<input type=password name=c> \
<input type=submit value=clear></form></pre>\n\
"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<h2>Change master pw</h2>\n\
<pre>\
<form action=sm method=get onsubmit=csm(this)>\n\
new pw:<input type=text name=n>\n\
old pw:<input type=password name=c> \
<input type=submit value=save></form>\n\
<hr>\n\
"));
        return(plen);
}

// this the key/master code login page
uint16_t print_webpage_login(char *actionurl)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a><br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=/p1.js></script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Enter passcode</h2>\n")); 
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action="));
        plen=fill_tcp_data(buf,plen,actionurl);
        plen=fill_tcp_data_p(buf,plen,PSTR(" method=get onsubmit=cy(this)>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<input type=hidden name=y>\n\
<input type=password name=c size=9> \n\
<input type=submit value=Enter></form>\n\
<br><hr>\n\
"));
        return(plen);
}

uint16_t print_webpage_key_free_control(void)
{
        uint16_t plen;
        int8_t i;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/sc>[set]</a> <a href=/hh>[history]</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=.>[refresh]</a><br>\n<pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("door states: |"));
        i=0;
        while(i<maxdoornum){
                gStrbuf[0]='.';
                // PC0, PC1, PC2, PC3:
                if (bit_is_clear(PINC,i)){
                        // closed
                        gStrbuf[0]='=';
                }
                i++;
                gStrbuf[1]='|';gStrbuf[2]='\0';
                plen=fill_tcp_data(buf,plen,gStrbuf);
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\n\n"));
        i=0;
        while(i<maxdoornum){
                itoa(i,gStrbuf,10); // convert integer to string
                plen=fill_tcp_data_p(buf,plen,PSTR("<form action=ad method=get>\n"));
                plen=fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=n value="));
                plen=fill_tcp_data(buf,plen,gStrbuf); // write i
                plen=fill_tcp_data_p(buf,plen,PSTR("><input type=submit value=\"activate door "));
                plen=fill_tcp_data(buf,plen,gStrbuf); // write i
                plen=fill_tcp_data_p(buf,plen,PSTR("\"></form>\n"));
                i++;
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("<hr>\n&copy;tuxgraphics\n"));
        return(plen);
}

// this the top level main web page
uint16_t print_webpage_main(void)
{
        uint16_t plen;
        int8_t i;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/sc>[set]</a> <a href=/kn>[keyFree]</a> <a href=/hh>[history]</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[refresh]</a><br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<script src=p1.js></script>\n<pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("door states: |"));
        i=0;
        while(i<maxdoornum){
                gStrbuf[0]='.';
                // PC0, PC1, PC2, PC3:
                if (bit_is_clear(PINC,i)){
                        // closed
                        gStrbuf[0]='=';
                }
                i++;
                gStrbuf[1]='|';gStrbuf[2]='\0';
                plen=fill_tcp_data(buf,plen,gStrbuf);
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=od method=get onsubmit=cy(this)>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<b>activate door</b> <select size=1 name=n>"));
        i=maxdoornum;
        while(i){
                i--;
                plen=fill_tcp_data_p(buf,plen,PSTR("<option>"));
                itoa(i,gStrbuf,10); // convert integer to string
                plen=fill_tcp_data(buf,plen,gStrbuf);
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("</select>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<input type=hidden name=y>\n\
<input type=password name=c size=9> \
<input type=submit value=Enter></form>\n\
<hr>\n&copy;tuxgraphics\n\
"));
        return(plen);
}

uint8_t keymatch(char *key)
{
        if (strlen(key)<2) return(0);
        if (strncmp(c1,key,sizeof(c1)-1)==0) return(1);
        if (strncmp(c2,key,sizeof(c2)-1)==0) return(2);
        return(0);
}

// key_in is in this format: random_arr_indexHEXNUMBERS
// return data in key_out. return value: 0 for failure and 1 for success. 
// key_out is the decoded string buffer which you must provide. max_key_out_len is the
// size (of the key_out string buffer). 
uint8_t decode_key(char *key_in,char *key_out,uint8_t max_key_out_len)
{
        uint8_t used_random_arr_index;
        uint8_t i=0;
        char hexstr[3];
        if (!isdigit(key_in[0])){
                return(0);
        }
        used_random_arr_index=*key_in - '0';
        key_in++; // skip over random_arr_index digit
        if (strlen(key_in)<2){ // must have at least two digits
                return(0);
        }
        hexstr[2]='\0';
        max_key_out_len--;
        while(*key_in && max_key_out_len){
                hexstr[0]=*key_in;
                key_in++;
                if (!*key_in) return(0); // must have even number of digits
                hexstr[1]=*key_in;
                key_out[i]=random_arr[used_random_arr_index] ^ strtol(hexstr,NULL,16);
                max_key_out_len--;
                key_in++;
                i++;
        }
        key_out[i]='\0';
        return(1);
}

// this is how Cookies work: http://www.nczonline.net/blog/2009/05/05/http-cookies-explained/
// We check here the user agent and any cookie that we might have
int8_t check_if_key_free_access_ok(char *str){
        int8_t i,j;
        char *s;
        uint8_t fp; // fingerprint
        s=strstr(str,"User-Agent:");
        fp=0;
        if (s){
                while( *s && *s!='\r' && *s!='\n'){
                        fp+=*s;  // fingerprint the string
                        s++;
                }
        }else{
                return(-1); // user agent not found
        }
        if (fp==0) fp=1; // do not use fp=0
        s=strstr(str,"Cookie:");
        if (!s) return(-1);
        str=strstr(s,"dk=");
        if (!str) return(-1);
        str+=3;  // start of cookie
        s=str;
        while(*str && *str!=';' && *str!=' ' && *str!='\r'){str++;}
        if (s==str) return(-1); // zero len not allowed
        *str='\0'; // terminate cookie string
        // check if we know this browser:
        i=0; 
        j=0;
        while(i<3){
                if (browser_fingerprint[i]==fp && strcmp(browser_cookie[i],s)==0){
                        j=1;
                        break;
                }
                i++;
        }
        if (j==0) return(-1);
        return(i); // i is now the index to browser_fingerprint/browser_cookie array
}

// takes a string of the form command/Number and analyse it (e.g "cc?tv=8%3A15&pw=secret HTTP/1.1")
// The first char of the url ('/') is already removed.
int8_t analyse_get_url(char *str)
{
        int8_t i,j;
        char *s;
#define KEY_BUF_SIZE 15
        char key[KEY_BUF_SIZE];
        uint8_t fp;
        // if you submit an empty form then android browsers send "? "
        if (str[0]==' ' || strncmp("? ",str,2)==0){ // end of url, main page
                if (check_if_key_free_access_ok(str)>=0){
                        gPlen=print_webpage_key_free_control();
                        return(10);
                }
                return(1); // normal main page with no key free access
        }
        // --------
        if (strncmp("favicon.ico",str,11)==0){
                gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("http://tuxgraphics.org/ico/gd.ico"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("\r\n\r\nContent-Type: text/html\r\n\r\n"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>301 Moved Permanently</h1>\n"));
                return(10);
        }
        // --------
        if (strncmp("p1.js",str,5)==0){
                gPlen=print_p1_js();
                return(10);
        }
        if (strncmp("p2.js",str,5)==0){
                gPlen=print_p2_js();
                return(10);
        }
        // --------
        if (strncmp("kn ",str,3)==0){ // allow key free access with cookie and fingerprint
                gPlen=print_webpage_nokey();
                browser_cookie_attempt++;
                return(10);
        }
        if (strncmp("kk?",str,3)==0){ // show set cookie ok page
                s=strstr(str,"User-Agent:");
                fp=0;
                if (s){
                        while(*s && *s!='\r' && *s!='\n'){
                                fp+=*s;  // fingerprint the string
                                s++;
                        }
                }
                if (fp==0) fp=1; // do not use fp=0
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"n")){
                        i=gStrbuf[0] - '0';
                        // the device number: 0..3
                        if (i < 0 || i >3){
                                return(-1); 
                        }
                }else{
                        return(0);
                }
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"y")){
                        urldecode(gStrbuf);
                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                if (strcmp(km,key)==0 && find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"e")){
                                        urldecode(gStrbuf); // e == expriry string
                                        store_hist(min_high,min_low,KM);
                                        browser_fingerprint[i]=fp; // i is the device number (0..2)
                                        // now make a random cookie string
                                        itoa(int_random|(browser_cookie_attempt<<8),browser_cookie[i],16);
                                        itoa(gsec,key,16);
                                        strcat(browser_cookie[i],key);

                                        gPlen=print_webpage_setcookie(gStrbuf,browser_cookie[i]);
                                        eeprom_store_browser_cookie();
                                        eeprom_store_browser_fingerprint();

                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        // --------
        if (strncmp("hh ",str,3)==0){ // code login for history
                if (check_if_key_free_access_ok(str)>=0){
                        gPlen=print_webpage_hh();
                        return(10);
                }
                gPlen=print_webpage_login("hm");
                return(10);
        }
        if (strncmp("hm",str,2)==0){ // show history after login
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"y")){
                        urldecode(gStrbuf);
                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                j=keymatch(key);
                                if (j){
                                        gPlen=print_webpage_hh();
                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        // --------
        if (strncmp("dd",str,2)==0){ // delete all key free access
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"y")){
                        urldecode(gStrbuf);
                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                if (strcmp(km,key)==0){
                                        store_hist(min_high,min_low,KM);
                                        i=0;
                                        while(i<3){
                                                // finger print =0 means no entry
                                                browser_fingerprint[i]=0;
                                                i++;
                                        }
                                        eeprom_store_browser_fingerprint();
                                        gPlen=print_webpage_ok();
                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        // --------
        if (strncmp("sc ",str,3)==0){ // master code login for setting
                gPlen=print_webpage_changekeys();
                return(10);
        }
        if (strncmp("sp",str,2)==0){ // show change keys page after master code login
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"y")){
                        urldecode(gStrbuf);
                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                if (strcmp(km,key)==0){
                                        i=0;
                                        if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"z1")){
                                                urldecode(gStrbuf);
                                                // empty fields are just ignored
                                                if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                                        strncpy(c1,key,sizeof(c1)-1);
                                                        c1[sizeof(c1)-1]='\0';
                                                        i++;
                                                }
                                        }
                                        if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"z2")){
                                                urldecode(gStrbuf);
                                                // empty fields are just ignored
                                                if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                                        strncpy(c2,key,sizeof(c2)-1);
                                                        c2[sizeof(c2)-1]='\0';
                                                        i++;
                                                }
                                        }
                                        eeprom_store_passwords();
                                        store_hist(min_high,min_low,KM);
                                        strcpy(gStrbuf,"no");
                                        if (i==1) strcpy(gStrbuf,"one");
                                        if (i==2) strcpy(gStrbuf,"two");
                                        gPlen=print_webpage_ok_pw_update(gStrbuf);
                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        // --------
        if (strncmp("sm",str,2)==0){ // save a new master key
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"y")){
                        urldecode(gStrbuf);
                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                if (strcmp(km,key)==0 && find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"z")){
                                        urldecode(gStrbuf);
                                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                                strncpy(km,key,sizeof(km)-1);
                                                km[sizeof(km)-1]='\0';
                                                store_hist(min_high,min_low,KM);
                                                eeprom_store_master();
                                                gPlen=print_webpage_ok();
                                                return(10);
                                        }
                                }
                        }
                }
                return(-1);
        }
        // --------
        // we can have "ad" or "od" here. It is important to use different form
        // names as some android browsers seem to cache forms based on that.
        if (str[1]=='d' && (str[0]=='o' || str[0]=='a')){ // open or close the door
                i=0; // door number
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"n")){
                        i=gStrbuf[0] - '0';
                        // the door number in "i" counts from 0. max 4 doors=0..3:
                        if (i < 0 || i >= maxdoornum){
                                return(-1); // not a valid door number
                        }
                }else{
                        return(0);
                }
                j=check_if_key_free_access_ok(str);
                if (j>=0){
                        // change door, PORTD7= door 0, PORTD6= door 1, ...
                        store_hist(min_high,min_low,(i<<4)|(j+4));// Cfree0=4 .. Cfree2=6
                        PORTD|=(1<<(7-i));
                        pushbuttontimer=50; // 1 sec relay on
                        gPlen=print_webpage_door_activatation(i);
                        return(10);
                }
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"y")){
                        urldecode(gStrbuf);
                        if (decode_key(gStrbuf,key,KEY_BUF_SIZE)){
                                j=keymatch(key);
                                if (j){
                                        // change door, PORTD7= door 0, PORTD6= door 1, ...
                                        store_hist(min_high,min_low,(i<<4)|(j+1));// C1=2 C2=3
                                        PORTD|=(1<<(7-i));
                                        pushbuttontimer=50; // 1 sec relay on
                                        gPlen=print_webpage_door_activatation(i);
                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        return(0);
}

// called when TCNT2==OCR2A, the clock ticks here
// that is in 50Hz intervals (49.8246Hz intervals, clock is a bit behind)
// This is used as a "clock" to store the data
ISR(TIMER2_COMPA_vect){
        static uint8_t cnt2step=0;
        cnt2step++;
        if (cnt2step>49){ 
                gsec++; // step sec
                cnt2step=0;
        }
        if (debounce){
                debounce--;
        }
        if (pushbuttontimer){
                pushbuttontimer--;
                if (pushbuttontimer==0){
                        ALLRELAYOFF;
                }
        }
}

/* setup timer T2 as an interrupt generating time base.
* You must call once sei() in the main program */
void init_cnt2(void)
{
        PRR&=~(1<<PRTIM2); // write power reduction register to zero
        TIMSK2=(1<<OCIE2A); // compare match on OCR2A
        TCNT2=0;  // init counter
        OCR2A=244; // value to compare against (we divide by 245)
        TCCR2A=(1<<WGM21); // do not change any output pin, clear at compare match
        // divide clock by 1024: 12.5MHz/1024=12207.0313 Hz
        TCCR2B=(1<<CS22)|(1<<CS21)|(1<<CS20); // clock divider, start counter
        // 12207.0313 / 245= 49.8246Hz
}


int main(void){
        uint16_t dat_p;
        uint8_t j=0;
        uint8_t gratutious_arp_cnt=0;
        uint8_t curr_doorstates=0;
        int8_t ledoff_after_sec=-1; // used to light the onboard LED for a few sec 
        int8_t i=0;
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        // enable PC0, PC1, ...,PC3 as input, door state
        DDRC&= ~(1<<DDC0)|(1<<DDC1)|(1<<DDC2)|(1<<DDC3);
        PORTC|= (1<<PINC0)|(1<<PINC1)|(1<<PINC2)|(1<<PINC3); // internal pullup resistor on, door switch goes to gnd
        // enable PD2 as input to reset master pw to default
        DDRD&= ~(1<<DDD2);
        PORTD|= (1<<PIND2); // internal pullup resistor on, to reset put to gnd during power on
        // LED
        DDRB|= (1<<DDB1); // enable PB1, LED as output
        LEDON;
        //initialize enc28j60
        enc28j60Init(mymac);
        LEDOFF;
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        //
        // time keeping
        init_cnt2();
        sei();

        // the transistor on PD7..PD4 (relay)
        DDRD|= (1<<DDD7)|(1<<DDD6)|(1<<DDD5)|(1<<DDD4);
        ALLRELAYOFF;
        // Magjack leds configuration, see enc28j60 datasheet, page 11 
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);

        curr_doorstates=PINC & 0xf;
        prev_doorstates=curr_doorstates;
        set_hist_endmarker(); // erase the history, we don't know how long we are power off and therefore not how to continue counting time.
        // restore permanent settings from eeprom if there are any:
        eeprom_read_passwords_condtional();
        eeprom_read_browser_fingerprint_conditional();
        eeprom_read_browser_cookie_conditional();
        // do not read master pw from eeprom if jumper is set during power on:
        if (bit_is_clear(PIND,PIND2)){
                eeprom_invalidate_master();
        }else{
                eeprom_read_master_conditional();
        }

        while(1){
                // get the next new packet:
                gPlen = enc28j60PacketReceive(BUFFER_SIZE, buf);
                if (gPlen){ // take entrophy from the network:
                        int_random++;
                }
                dat_p=packetloop_icmp_tcp(buf,gPlen);

                //dat_p will ne unequal to zero if there is a valid packet (without crc error) 
                if(dat_p==0){
                        // no packets, check door states, debounce 
                        // is important for a mechanical switch.
                        curr_doorstates=PINC & 0xf;
                        j=(curr_doorstates^prev_doorstates );
                        if (j && debounce==0){
                                i=4;
                                while(i){
                                        i--;
                                        if( j & (1<<i)){
                                                if (curr_doorstates & (1<<i)){
                                                        // door "i" open
                                                        store_hist(min_high,min_low,i<<4);
                                                }else{
                                                        // door "i" close
                                                        store_hist(min_high,min_low,(i<<4) | 1);
                                                }
                                        }
                                }
                                // one of the doors changed
                                LEDON;
                                // switch LED off in 2 sec 
                                ledoff_after_sec=(gsec+2)%60;
                                prev_doorstates=curr_doorstates;
                                debounce=10; // 1/4sec debounce time
                        }
                        // Make this work with broken wlan/wifi equipment.
                        // Some of those wifi bridges or wifi client APs
                        // need once in a while a packet otherwise they
                        // forget the MAC address of the connected device.
                        if (gsec==7 && gratutious_arp_cnt==0){
                                gratutious_arp_cnt=5; // every 5min
                                gratutious_arp(buf);
                        }
                        // no packets, do time keeping
                        //
                        // if the crystal would be exact then we
                        // would be 303.0609 seconds behind every day.
                        // 24*3600 - 24*3600 * 0.9965=303.0609
                        // We compensate one min  every 288min :
                        // 60*288 - 60*288*0.9965 + 60==17280-17219.52+60=0.48
                        // That is we are now 0.48sec behind every 288min
                        // We compensate every 7500 min (125h) by another
                        // min and the clock should be accurate.
                        if (ledoff_after_sec>=0){ // -1 means function is disabeled
                                if (gsec==ledoff_after_sec){
                                        LEDOFF;
                                        ledoff_after_sec=-1;
                                }
                        }
                        if (gsec>59){ // sec
                                gsec=0;
                                min_low++; // count up min
                                // count gratutious_arp_cnt down every min:
                                if (gratutious_arp_cnt) gratutious_arp_cnt--;
                                if (min_low==0){ // wrap
                                        min_high++;
                                }
                                if (min_low%288==0){
                                        min_low++;
                                        if (min_low==0){ // wrap
                                                min_high++;
                                        }
                                }
                                if (min_low%7500==0){
                                        min_low++;
                                        if (min_low==0){ // wrap
                                                min_high++;
                                        }
                                }
                                // store a new random number and rotate the index:
                                random_arr_index=(random_arr_index +1) % 3;
                                random_arr[random_arr_index]=int_random;
                        }
                        //
                        continue;
                }
                buf[gPlen-1]='\0'; // make sure the http part of the buffer is terminated
                // analyse the url 
                // move one char ahead to remove the initial "/":
                i=analyse_get_url((char *)&(buf[dat_p+5]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (i==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1><a href=/>continue</a>\n"));
                        goto SENDTCP;
                }
                if (i==0){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                        goto SENDTCP;
                }
                if (i==10){
                        goto SENDTCP;
                }
                // just display the status (i==1):
                gPlen=print_webpage_main();
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send web page data
                // tcp port www end
                //
        }
        return (0);
}
// -- end of main
